/**
 ** sahandCntrlTaskCodels.cc
 **
 ** Codels used by the control task sahandCntrlTask
 **
 ** Author: 
 ** Date: 
 **
 **/

#include <portLib.h>
#include <h2mathLib.h>

#include "server/sahandHeader.h"
#include "sahandConst.h"
#include "sahandPrivate.h"

static STATUS
sahandControlPort(int port, int *report)
{
	
	if (port < SAH_PORT_1 || port > SAH_PORT_2) {
		*report = S_sahand_BAD_PARAMETERS;
		return ERROR;
	}
	if (!SAHAND_PORT_ACTIVE(port)) {
		*report = S_sahand_INACTIVE_PORT;
		return ERROR;
	}
	return OK;
}

/*------------------------------------------------------------------------
 * sahandSetControllerControl  -  control codel of CONTROL request SetController
 *
 * Description: check the validity of the controller type
 *              and register the command to be executed
 * Report: OK
 *                 S_sahand_BAD_PARAMETERS
 *                 S_sahand_INACTIVE_PORT
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandSetControllerControl(SAHAND_THIS_CONTROLLER_TYPE *sah_controller, int *report)
{
	int port = sah_controller->stct_port;
	SAHAND_CONTROLLER_TYPE controller = sah_controller->stct_controller;
	
	if (sahandControlPort(port, report) == ERROR)
		return ERROR;
	if (controller < SAH_CONTROLLER_RESET ||
	    controller > SAH_CONTROLLER_IMPEDANCE) {
		*report =  S_sahand_BAD_PARAMETERS;
		return ERROR;
	}
	SDI_F->sah_params[port].sap_controller = controller;
	SDI_F->sah_command[port] |= SAH_COMMAND_SET_CONTROLLER;
	return OK;
}

/*------------------------------------------------------------------------
 * sahandMoveFingerControl  -  control codel of CONTROL request MoveFinger
 *
 * Description: check bounds of finger move commands
 *              and register the command to be executed
 * Report: OK
 *                 S_sahand_BAD_PARAMETERS
 *                 S_sahand_INACTIVE_PORT
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandMoveFingerControl(SAHAND_THIS_FINGER_REF *sah_finger_ref, int *report)
{
	double *angle, *vel;
	SAHAND_FINGER_REF *ref;
	int port = sah_finger_ref->stfr_port;
	
	ref = &SDI_F->sah_ref[port].sr_fingers[0];
	if (sahandControlPort(port, report) == ERROR)
		return ERROR;
	angle = &sah_finger_ref->stfr_finger.sfr_angle[0];
	vel = &sah_finger_ref->stfr_finger.sfr_vel[0];
	/* Check finger Id  */
	if (sah_finger_ref->stfr_id < 0 ||
	    sah_finger_ref->stfr_id >= SAH_MAX_FINGERS) {
		*report = S_sahand_BAD_PARAMETERS;
		return ERROR;
	}
	/* Check values */
	if (sahandFingerPositionCheck(angle)) {
	  *report = S_sahand_BAD_PARAMETERS;
	  return ERROR;
	}
	if (sahandFingerVelocityCheck(vel)) {
	  *report = S_sahand_BAD_PARAMETERS;
	  return ERROR;
	}
	/* copy data to finger struct */
	memcpy(&ref[sah_finger_ref->stfr_id],
	    &sah_finger_ref->stfr_finger,
	    sizeof(SAHAND_FINGER_REF));

	SDI_F->sah_command[port] |= SAH_COMMAND_MOVE_HAND;
	return OK;
}

/*------------------------------------------------------------------------
 * sahandSetStiffnessControl  -  control codel of CONTROL request SetStiffness
 *
 * Description: checks bounds of the stiffness factor command.
 *              and register the command to be executed
 * Report: OK
 *                 S_sahand_BAD_PARAMETERS
 *                 S_sahand_INACTIVE_PORT
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandSetStiffnessControl(SAHAND_THIS_FINGER_STIFFNESS *sah_finger_stiffness, int *report)
{
	double *value = &sah_finger_stiffness->stfs_finger.ssf_value[0];
	int i, finger = sah_finger_stiffness->stfs_id,
	    port = sah_finger_stiffness->stfs_port;

	if (sahandControlPort(port, report) == ERROR)
		return ERROR;
	/* Check finger Id  */
	if (finger < 0 || finger >= SAH_MAX_FINGERS) {
		*report = S_sahand_BAD_PARAMETERS;
		return ERROR;
	}
	/* Check values */
	if (sahandFingerStiffnessCheck(value)) {
	  *report = S_sahand_BAD_PARAMETERS;
	  return ERROR;
	}
	/* copy data */
	for (i = 0; i < 3; i++) 
		SDI_F->sah_params[port].sap_stiffness[finger].ssf_value[i] = 
		    value[i];
	SDI_F->sah_command[port] |= SAH_COMMAND_SET_STIFFNESS_FACTOR;
	return OK;
}

/*------------------------------------------------------------------------
 * sahandRotateThumbControl  -  control codel of CONTROL request RotateThumb
 *
 * Description: check bounds of the thumb angle command
 *              and register the command to be executed
 * Report: OK
 *                 S_sahand_BAD_PARAMETERS
 *                 S_sahand_INACTIVE_PORT
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandRotateThumbControl(SAHAND_THIS_THUMB_ANGLE *sah_thumb_angle, int *report)
{
  int port = sah_thumb_angle->stta_port;
  double angle = sah_thumb_angle->stta_angle;
  
  if (sahandControlPort(port, report) == ERROR)
    return ERROR;
  if (angle < SAH_MIN_THUMB_ORIENTATION ||
      angle > SAH_MAX_THUMB_ORIENTATION) {
    *report = S_sahand_BAD_PARAMETERS;
    return ERROR;
  }
  SDI_F->sah_ref[port].sr_thumb_orientation = angle;
  SDI_F->sah_command[port] |= SAH_COMMAND_SET_THUMB_ANGLE;
  return OK;
}

/*------------------------------------------------------------------------
 * sahandSetHandStiffnessCntrl  -  control codel of CONTROL request SetHandStiffness
 *
 * Description:    
 * Report: OK
 *                 S_sahand_BAD_HAND
 *                 S_sahand_BAD_STIFFNESS
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandSetHandStiffnessCntrl(SAHAND_HAND_STIFFNESS *hand_stiffness, int *report)
{
  int port = hand_stiffness->hand;

  if (sahandControlPort(port, report) == ERROR)
    return ERROR;
  SDI_F->sah_params[port].sap_stiffness[SAH_FINGER_THUMB].ssf_value[SAHAND_DISTAL] = hand_stiffness->thumb.distal;
  SDI_F->sah_params[port].sap_stiffness[SAH_FINGER_THUMB].ssf_value[SAHAND_PROXIMAL] = hand_stiffness->thumb.proximal;
  SDI_F->sah_params[port].sap_stiffness[SAH_FINGER_THUMB].ssf_value[SAHAND_BASE] = hand_stiffness->thumb.base;
  SDI_F->sah_params[port].sap_stiffness[SAH_FINGER_FIRST].ssf_value[SAHAND_DISTAL] = hand_stiffness->first.distal;
  SDI_F->sah_params[port].sap_stiffness[SAH_FINGER_FIRST].ssf_value[SAHAND_PROXIMAL] = hand_stiffness->first.proximal;
  SDI_F->sah_params[port].sap_stiffness[SAH_FINGER_FIRST].ssf_value[SAHAND_BASE] = hand_stiffness->first.base;
  SDI_F->sah_params[port].sap_stiffness[SAH_FINGER_MIDDLE].ssf_value[SAHAND_DISTAL] = hand_stiffness->middle.distal;
  SDI_F->sah_params[port].sap_stiffness[SAH_FINGER_MIDDLE].ssf_value[SAHAND_PROXIMAL] = hand_stiffness->middle.proximal;
  SDI_F->sah_params[port].sap_stiffness[SAH_FINGER_MIDDLE].ssf_value[SAHAND_BASE] = hand_stiffness->middle.base;
  SDI_F->sah_params[port].sap_stiffness[SAH_FINGER_RING].ssf_value[SAHAND_DISTAL] = hand_stiffness->ring.distal;
  SDI_F->sah_params[port].sap_stiffness[SAH_FINGER_RING].ssf_value[SAHAND_PROXIMAL] = hand_stiffness->ring.proximal;
  SDI_F->sah_params[port].sap_stiffness[SAH_FINGER_RING].ssf_value[SAHAND_BASE] = hand_stiffness->ring.base;
  SDI_F->sah_command[port] |= SAH_COMMAND_SET_STIFFNESS_FACTOR;
  return OK;
}

/*------------------------------------------------------------------------
 * sahandGetHandStiffnessCntrl  -  control codel of CONTROL request GetHandStiffness
 *
 * Description:    
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandGetHandStiffnessCntrl(SAHAND_PORT *hand, int *report)
{
  if (sahandControlPort(*hand, report) == ERROR)
    return ERROR;
  SDI_F->hand_stiffness.thumb.distal    = SDI_F->sah_params[*hand].sap_stiffness[SAH_FINGER_THUMB].ssf_value[SAHAND_DISTAL];
  SDI_F->hand_stiffness.thumb.proximal  = SDI_F->sah_params[*hand].sap_stiffness[SAH_FINGER_THUMB].ssf_value[SAHAND_PROXIMAL];
  SDI_F->hand_stiffness.thumb.base      = SDI_F->sah_params[*hand].sap_stiffness[SAH_FINGER_THUMB].ssf_value[SAHAND_BASE];
  SDI_F->hand_stiffness.first.distal    = SDI_F->sah_params[*hand].sap_stiffness[SAH_FINGER_FIRST].ssf_value[SAHAND_DISTAL];
  SDI_F->hand_stiffness.first.proximal  = SDI_F->sah_params[*hand].sap_stiffness[SAH_FINGER_FIRST].ssf_value[SAHAND_PROXIMAL];
  SDI_F->hand_stiffness.first.base      = SDI_F->sah_params[*hand].sap_stiffness[SAH_FINGER_FIRST].ssf_value[SAHAND_BASE];
  SDI_F->hand_stiffness.middle.distal   = SDI_F->sah_params[*hand].sap_stiffness[SAH_FINGER_MIDDLE].ssf_value[SAHAND_DISTAL];  
  SDI_F->hand_stiffness.middle.proximal = SDI_F->sah_params[*hand].sap_stiffness[SAH_FINGER_MIDDLE].ssf_value[SAHAND_PROXIMAL];
  SDI_F->hand_stiffness.middle.base     = SDI_F->sah_params[*hand].sap_stiffness[SAH_FINGER_MIDDLE].ssf_value[SAHAND_BASE];
  SDI_F->hand_stiffness.ring.distal     = SDI_F->sah_params[*hand].sap_stiffness[SAH_FINGER_RING].ssf_value[SAHAND_DISTAL];  
  SDI_F->hand_stiffness.ring.proximal   = SDI_F->sah_params[*hand].sap_stiffness[SAH_FINGER_RING].ssf_value[SAHAND_PROXIMAL];
  SDI_F->hand_stiffness.ring.base       = SDI_F->sah_params[*hand].sap_stiffness[SAH_FINGER_RING].ssf_value[SAHAND_BASE];
  return OK;
}

/*------------------------------------------------------------------------
 * sahandGetHandStiffnessMinCntrl  -  control codel of CONTROL request GetHandStiffnessMin
 *
 * Description:    
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandGetHandStiffnessMinCntrl(int *report)
{
  SDI_F->hand_stiffness.thumb.distal    = 0.0;
  SDI_F->hand_stiffness.thumb.proximal  = 0.0;
  SDI_F->hand_stiffness.thumb.base      = 0.0;
  SDI_F->hand_stiffness.first.distal    = 0.0;
  SDI_F->hand_stiffness.first.proximal  = 0.0;
  SDI_F->hand_stiffness.first.base      = 0.0;
  SDI_F->hand_stiffness.middle.distal   = 0.0;
  SDI_F->hand_stiffness.middle.proximal = 0.0;
  SDI_F->hand_stiffness.middle.base     = 0.0;
  SDI_F->hand_stiffness.ring.distal     = 0.0;
  SDI_F->hand_stiffness.ring.proximal   = 0.0;
  SDI_F->hand_stiffness.ring.base       = 0.0;
  return OK;
}

/*------------------------------------------------------------------------
 * sahandGetHandStiffnessMaxCntrl  -  control codel of CONTROL request GetHandStiffnessMax
 *
 * Description:    
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandGetHandStiffnessMaxCntrl(int *report)
{
  SDI_F->hand_stiffness.thumb.distal    = SAH_MAX_DISTAL_STIFFNESS;
  SDI_F->hand_stiffness.thumb.proximal  = SAH_MAX_PROXIMAL_STIFFNESS;
  SDI_F->hand_stiffness.thumb.base      = SAH_MAX_BASE_STIFFNESS;
  SDI_F->hand_stiffness.first.distal    = SAH_MAX_DISTAL_STIFFNESS;
  SDI_F->hand_stiffness.first.proximal  = SAH_MAX_PROXIMAL_STIFFNESS;
  SDI_F->hand_stiffness.first.base      = SAH_MAX_BASE_STIFFNESS;
  SDI_F->hand_stiffness.middle.distal   = SAH_MAX_DISTAL_STIFFNESS;
  SDI_F->hand_stiffness.middle.proximal = SAH_MAX_PROXIMAL_STIFFNESS;
  SDI_F->hand_stiffness.middle.base     = SAH_MAX_BASE_STIFFNESS;
  SDI_F->hand_stiffness.ring.distal     = SAH_MAX_DISTAL_STIFFNESS;
  SDI_F->hand_stiffness.ring.proximal   = SAH_MAX_PROXIMAL_STIFFNESS;
  SDI_F->hand_stiffness.ring.base       = SAH_MAX_BASE_STIFFNESS;
  return OK;
}

/*------------------------------------------------------------------------
 * sahandSetHandVelocityCntrl  -  control codel of CONTROL request SetHandVelocity
 *
 * Description:    
 * Report: OK
 *                 S_sahand_BAD_HAND
 *                 S_sahand_BAD_VELOCITY
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandSetHandVelocityCntrl(SAHAND_HAND_VELOCITY *hand_velocity, int *report)
{
  int port = hand_velocity->hand;
  double vel[3];

  if (sahandControlPort(port, report) == ERROR)
    return ERROR;
  vel[SAHAND_DISTAL]   = hand_velocity->thumb.distal;	
  vel[SAHAND_PROXIMAL] = hand_velocity->thumb.proximal;	
  vel[SAHAND_BASE]     = hand_velocity->thumb.base;		
  if (sahandFingerVelocityCheck(vel)) {
    *report = S_sahand_BAD_PARAMETERS;
    return ERROR;
  }
  vel[SAHAND_DISTAL]   = hand_velocity->first.distal;	
  vel[SAHAND_PROXIMAL] = hand_velocity->first.proximal;	
  vel[SAHAND_BASE]     = hand_velocity->first.base;		
  if (sahandFingerVelocityCheck(vel)) {
    *report = S_sahand_BAD_PARAMETERS;
    return ERROR;
  }
  vel[SAHAND_DISTAL]   = hand_velocity->middle.distal;	
  vel[SAHAND_PROXIMAL] = hand_velocity->middle.proximal;	
  vel[SAHAND_BASE]     = hand_velocity->middle.base;		
  if (sahandFingerVelocityCheck(vel)) {
    *report = S_sahand_BAD_PARAMETERS;
    return ERROR;
  }
  vel[SAHAND_DISTAL]   = hand_velocity->ring.distal;		
  vel[SAHAND_PROXIMAL] = hand_velocity->ring.proximal;	
  vel[SAHAND_BASE]     = hand_velocity->ring.base;             
  if (sahandFingerVelocityCheck(vel)) {
    *report = S_sahand_BAD_PARAMETERS;
    return ERROR;
  }
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_THUMB].sfr_vel[SAHAND_DISTAL] = hand_velocity->thumb.distal;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_THUMB].sfr_vel[SAHAND_PROXIMAL] = hand_velocity->thumb.proximal;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_THUMB].sfr_vel[SAHAND_BASE] = hand_velocity->thumb.base;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_FIRST].sfr_vel[SAHAND_DISTAL] = hand_velocity->first.distal;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_FIRST].sfr_vel[SAHAND_PROXIMAL] = hand_velocity->first.proximal;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_FIRST].sfr_vel[SAHAND_BASE] = hand_velocity->first.base;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_MIDDLE].sfr_vel[SAHAND_DISTAL] = hand_velocity->middle.distal;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_MIDDLE].sfr_vel[SAHAND_PROXIMAL] = hand_velocity->middle.proximal;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_MIDDLE].sfr_vel[SAHAND_BASE] = hand_velocity->middle.base;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_RING].sfr_vel[SAHAND_DISTAL] = hand_velocity->ring.distal;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_RING].sfr_vel[SAHAND_PROXIMAL] = hand_velocity->ring.proximal;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_RING].sfr_vel[SAHAND_BASE] = hand_velocity->ring.base;
  SDI_F->sah_command[port] |= SAH_COMMAND_MOVE_HAND;
  return OK;
}

/*------------------------------------------------------------------------
 * sahandGetHandVelocityCntrl  -  control codel of CONTROL request GetHandVelocity
 *
 * Description:    
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandGetHandVelocityCntrl(SAHAND_PORT *hand, int *report)
{
  if (sahandControlPort(*hand, report) == ERROR)
    return ERROR;
  SDI_F->hand_velocity.thumb.distal    = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_THUMB].sfs_vel[SAHAND_DISTAL];
  SDI_F->hand_velocity.thumb.proximal  = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_THUMB].sfs_vel[SAHAND_PROXIMAL];
  SDI_F->hand_velocity.thumb.base      = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_THUMB].sfs_vel[SAHAND_BASE];
  SDI_F->hand_velocity.first.distal    = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_FIRST].sfs_vel[SAHAND_DISTAL];
  SDI_F->hand_velocity.first.proximal  = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_FIRST].sfs_vel[SAHAND_PROXIMAL];
  SDI_F->hand_velocity.first.base      = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_FIRST].sfs_vel[SAHAND_BASE];
  SDI_F->hand_velocity.middle.distal   = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_MIDDLE].sfs_vel[SAHAND_DISTAL];	
  SDI_F->hand_velocity.middle.proximal = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_MIDDLE].sfs_vel[SAHAND_PROXIMAL];
  SDI_F->hand_velocity.middle.base     = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_MIDDLE].sfs_vel[SAHAND_BASE];
  SDI_F->hand_velocity.ring.distal     = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_RING].sfs_vel[SAHAND_DISTAL];	
  SDI_F->hand_velocity.ring.proximal   = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_RING].sfs_vel[SAHAND_PROXIMAL];
  SDI_F->hand_velocity.ring.base       = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_RING].sfs_vel[SAHAND_BASE];
  return OK;
}

/*------------------------------------------------------------------------
 * sahandGetHandVelocityMaxCntrl  -  control codel of CONTROL request GetHandVelocityMax
 *
 * Description:    
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandGetHandVelocityMaxCntrl(int *report)
{
  SDI_F->hand_stiffness.thumb.distal    = SAH_MAX_DISTAL_VEL;
  SDI_F->hand_stiffness.thumb.proximal  = SAH_MAX_PROXIMAL_VEL;
  SDI_F->hand_stiffness.thumb.base      = SAH_MAX_BASE_VEL;
  SDI_F->hand_stiffness.first.distal    = SAH_MAX_DISTAL_VEL;
  SDI_F->hand_stiffness.first.proximal  = SAH_MAX_PROXIMAL_VEL;
  SDI_F->hand_stiffness.first.base      = SAH_MAX_BASE_VEL;
  SDI_F->hand_stiffness.middle.distal   = SAH_MAX_DISTAL_VEL;
  SDI_F->hand_stiffness.middle.proximal = SAH_MAX_PROXIMAL_VEL;
  SDI_F->hand_stiffness.middle.base     = SAH_MAX_BASE_VEL;
  SDI_F->hand_stiffness.ring.distal     = SAH_MAX_DISTAL_VEL;
  SDI_F->hand_stiffness.ring.proximal   = SAH_MAX_PROXIMAL_VEL;
  SDI_F->hand_stiffness.ring.base       = SAH_MAX_BASE_VEL;
  return OK;
}

/*------------------------------------------------------------------------
 * sahandGetHandVelocityMinCntrl  -  control codel of CONTROL request GetHandVelocityMin
 *
 * Description:    
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandGetHandVelocityMinCntrl(int *report)
{
  SDI_F->hand_stiffness.thumb.distal    = 0.0;
  SDI_F->hand_stiffness.thumb.proximal  = 0.0;
  SDI_F->hand_stiffness.thumb.base      = 0.0;
  SDI_F->hand_stiffness.first.distal    = 0.0;
  SDI_F->hand_stiffness.first.proximal  = 0.0;
  SDI_F->hand_stiffness.first.base      = 0.0;
  SDI_F->hand_stiffness.middle.distal   = 0.0;
  SDI_F->hand_stiffness.middle.proximal = 0.0;
  SDI_F->hand_stiffness.middle.base     = 0.0;
  SDI_F->hand_stiffness.ring.distal     = 0.0;
  SDI_F->hand_stiffness.ring.proximal   = 0.0;
  SDI_F->hand_stiffness.ring.base       = 0.0;
  return OK;
}

/*------------------------------------------------------------------------
 * sahandHandMoveCntrl  -  control codel of CONTROL request HandMove
 *
 * Description:    
 * Report: OK
 *                 S_sahand_BAD_HAND
 *                 S_sahand_BAD_POSITION
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandHandMoveCntrl(SAHAND_HAND_POSITION *hand_position, int *report)
{
  int port = hand_position->hand;
  double angle[3];

  if (sahandControlPort(port, report) == ERROR)
    return ERROR;
  angle[SAHAND_DISTAL]   = hand_position->thumb.distal;	
  angle[SAHAND_PROXIMAL] = hand_position->thumb.proximal;	
  angle[SAHAND_BASE]     = hand_position->thumb.base;		
  if (sahandFingerPositionCheck(angle)) {
    *report = S_sahand_BAD_PARAMETERS;
    return ERROR;
  }
  angle[SAHAND_DISTAL]   = hand_position->first.distal;	
  angle[SAHAND_PROXIMAL] = hand_position->first.proximal;	
  angle[SAHAND_BASE]     = hand_position->first.base;		
  if (sahandFingerPositionCheck(angle)) {
    *report = S_sahand_BAD_PARAMETERS;
    return ERROR;
  }
  angle[SAHAND_DISTAL]   = hand_position->middle.distal;	
  angle[SAHAND_PROXIMAL] = hand_position->middle.proximal;	
  angle[SAHAND_BASE]     = hand_position->middle.base;		
  if (sahandFingerPositionCheck(angle)) {
    *report = S_sahand_BAD_PARAMETERS;
    return ERROR;
  }
  angle[SAHAND_DISTAL]   = hand_position->ring.distal;		
  angle[SAHAND_PROXIMAL] = hand_position->ring.proximal;	
  angle[SAHAND_BASE]     = hand_position->ring.base;             
  if (sahandFingerPositionCheck(angle)) {
    *report = S_sahand_BAD_PARAMETERS;
    return ERROR;
  }
  if (hand_position->thumb_orientation < SAH_MIN_THUMB_ORIENTATION ||
      hand_position->thumb_orientation > SAH_MAX_THUMB_ORIENTATION) {
    *report = S_sahand_BAD_PARAMETERS;
    return ERROR;
  }
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_THUMB].sfr_angle[SAHAND_DISTAL] = hand_position->thumb.distal;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_THUMB].sfr_angle[SAHAND_PROXIMAL] = hand_position->thumb.proximal;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_THUMB].sfr_angle[SAHAND_BASE] = hand_position->thumb.base;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_FIRST].sfr_angle[SAHAND_DISTAL] = hand_position->first.distal;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_FIRST].sfr_angle[SAHAND_PROXIMAL] = hand_position->first.proximal;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_FIRST].sfr_angle[SAHAND_BASE] = hand_position->first.base;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_MIDDLE].sfr_angle[SAHAND_DISTAL] = hand_position->middle.distal;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_MIDDLE].sfr_angle[SAHAND_PROXIMAL] = hand_position->middle.proximal;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_MIDDLE].sfr_angle[SAHAND_BASE] = hand_position->middle.base;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_RING].sfr_angle[SAHAND_DISTAL] = hand_position->ring.distal;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_RING].sfr_angle[SAHAND_PROXIMAL] = hand_position->ring.proximal;
  SDI_F->sah_ref[port].sr_fingers[SAH_FINGER_RING].sfr_angle[SAHAND_BASE] = hand_position->ring.base;
  SDI_F->sah_ref[port].sr_thumb_orientation = hand_position->thumb_orientation;
  SDI_F->sah_command[port] |= SAH_COMMAND_MOVE_HAND;
  return OK;
}

/*------------------------------------------------------------------------
 * sahandGetHandPositionCntrl  -  control codel of CONTROL request GetHandPosition
 *
 * Description:    
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandGetHandPositionCntrl(SAHAND_PORT *hand, int *report)
{
  if (sahandControlPort(*hand, report) == ERROR)
    return ERROR;
  SDI_F->hand_position.hand = *hand;
  SDI_F->hand_position.thumb.distal    = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_THUMB].sfs_pos[SAHAND_DISTAL];
  SDI_F->hand_position.thumb.proximal  = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_THUMB].sfs_pos[SAHAND_PROXIMAL];
  SDI_F->hand_position.thumb.base      = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_THUMB].sfs_pos[SAHAND_BASE];
  SDI_F->hand_position.first.distal    = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_FIRST].sfs_pos[SAHAND_DISTAL];
  SDI_F->hand_position.first.proximal  = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_FIRST].sfs_pos[SAHAND_PROXIMAL];
  SDI_F->hand_position.first.base      = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_FIRST].sfs_pos[SAHAND_BASE];
  SDI_F->hand_position.middle.distal   = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_MIDDLE].sfs_pos[SAHAND_DISTAL];	
  SDI_F->hand_position.middle.proximal = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_MIDDLE].sfs_pos[SAHAND_PROXIMAL];
  SDI_F->hand_position.middle.base     = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_MIDDLE].sfs_pos[SAHAND_BASE];
  SDI_F->hand_position.ring.distal     = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_RING].sfs_pos[SAHAND_DISTAL];	
  SDI_F->hand_position.ring.proximal   = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_RING].sfs_pos[SAHAND_PROXIMAL];
  SDI_F->hand_position.ring.base       = SDI_F->sah_state[*hand].ss_finger[SAH_FINGER_RING].sfs_pos[SAHAND_BASE];
  SDI_F->hand_position.thumb_orientation = SDI_F->sah_state[*hand].ss_thumb.sts_angle;
  return OK;
}

/*------------------------------------------------------------------------
 * sahandGetHandPositionMinCntrl  -  control codel of CONTROL request GetHandPositionMin
 *
 * Description:    
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandGetHandPositionMinCntrl(int *report)
{
  SDI_F->hand_position.thumb.distal    = SAH_MIN_DISTAL_ANGLE;
  SDI_F->hand_position.thumb.proximal  = SAH_MIN_PROXIMAL_ANGLE;
  SDI_F->hand_position.thumb.base      = SAH_MIN_BASE_ANGLE;
  SDI_F->hand_position.first.distal    = SAH_MIN_DISTAL_ANGLE;
  SDI_F->hand_position.first.proximal  = SAH_MIN_PROXIMAL_ANGLE;
  SDI_F->hand_position.first.base      = SAH_MIN_BASE_ANGLE;
  SDI_F->hand_position.middle.distal   = SAH_MIN_DISTAL_ANGLE;
  SDI_F->hand_position.middle.proximal = SAH_MIN_PROXIMAL_ANGLE;
  SDI_F->hand_position.middle.base     = SAH_MIN_BASE_ANGLE;
  SDI_F->hand_position.ring.distal     = SAH_MIN_DISTAL_ANGLE;
  SDI_F->hand_position.ring.proximal   = SAH_MIN_PROXIMAL_ANGLE;
  SDI_F->hand_position.ring.base       = SAH_MIN_BASE_ANGLE;
  return OK;
}

/*------------------------------------------------------------------------
 * sahandGetHandPositionMaxCntrl  -  control codel of CONTROL request GetHandPositionMax
 *
 * Description:    
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandGetHandPositionMaxCntrl(int *report)
{
  SDI_F->hand_position.thumb.distal    = SAH_MAX_DISTAL_ANGLE;
  SDI_F->hand_position.thumb.proximal  = SAH_MAX_PROXIMAL_ANGLE;
  SDI_F->hand_position.thumb.base      = SAH_MAX_BASE_ANGLE;
  SDI_F->hand_position.first.distal    = SAH_MAX_DISTAL_ANGLE;
  SDI_F->hand_position.first.proximal  = SAH_MAX_PROXIMAL_ANGLE;
  SDI_F->hand_position.first.base      = SAH_MAX_BASE_ANGLE;
  SDI_F->hand_position.middle.distal   = SAH_MAX_DISTAL_ANGLE;
  SDI_F->hand_position.middle.proximal = SAH_MAX_PROXIMAL_ANGLE;
  SDI_F->hand_position.middle.base     = SAH_MAX_BASE_ANGLE;
  SDI_F->hand_position.ring.distal     = SAH_MAX_DISTAL_ANGLE;
  SDI_F->hand_position.ring.proximal   = SAH_MAX_PROXIMAL_ANGLE;
  SDI_F->hand_position.ring.base       = SAH_MAX_BASE_ANGLE;
  return OK;
}

/*------------------------------------------------------------------------
 * sahandGraspBottleCntrl  -  control codel of EXEC request GraspBottle
 *
 * Description:    
 * Report: OK
 *                 S_sahand_GRASP_ERROR
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandGraspBottleCntrl(int *report)
{
  /* ... add your code here ... */
  return OK;
}

/*------------------------------------------------------------------------
 * sahandDefineGraspCntrl  -  control codel of CONTROL request DefineGrasp
 *
 * Description:    
 * Report: OK
 *                 S_sahand_BAD_HAND
 *                 S_sahand_BAD_FINGER
 *                 S_sahand_BAD_POSITION
 *                 S_sahand_BAD_STIFFNESS
 *                 S_sahand_BAD_VELOCITY
 *                 S_sahand_BAD_FORCE
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandDefineGraspCntrl(SAHAND_CLOSE_SET *sah_close_set, int *report)
{
  SAHAND_PORT port = sah_close_set->port;
  SAHAND_FINGER finger = sah_close_set->finger;

  if (sahandControlPort(port, report) == ERROR)
    return ERROR;
  if (finger < SAH_FINGER_THUMB || finger > SAH_FINGER_RING) {
    *report = S_sahand_BAD_FINGER;
    return ERROR;
  }
  if (sahandFingerPositionCheck(&(sah_close_set->data.positionEnd[0]))) {
    *report = S_sahand_BAD_POSITION;
    return ERROR;
  }
  // sah_close_set->data.positionDelta : limited at use time
  if (sahandFingerStiffnessCheck(&(sah_close_set->data.stiffness[0])) ||
      sahandFingerStiffnessCheck(&(sah_close_set->data.stiffnessClose[0]))) {
    *report = S_sahand_BAD_STIFFNESS;
    return ERROR;
  }
  if (sahandFingerVelocityCheck(&(sah_close_set->data.velocity[0])) ||
      sahandFingerVelocityCheck(&(sah_close_set->data.velocityClose[0]))) {
    *report = S_sahand_BAD_VELOCITY;
    return ERROR;
  }
  // sah_close_set->data.force not tested, as force just define limit in 
  //   sahand module.
  SDI_F->sah_close[port].fingers[finger] = sah_close_set->data;
  return OK;
}

/*------------------------------------------------------------------------
 * sahandGraspGetParaCntrl  -  control codel of CONTROL request GraspGetPara
 *
 * Description:    
 * Report: OK
 *                 S_sahand_BAD_HAND
 *                 S_sahand_BAD_FINGER
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandGraspGetParaCntrl(SAHAND_CLOSE_FINGER *sah_close_finger, int *report)
{
  SAHAND_PORT port = sah_close_finger->port;
  SAHAND_FINGER finger = sah_close_finger->finger;

  if (sahandControlPort(port, report) == ERROR)
    return ERROR;
  if (finger < SAH_FINGER_THUMB || finger > SAH_FINGER_RING) {
    *report = S_sahand_BAD_FINGER;
    return ERROR;
  }
  SDI_F->sah_close_set.port = port;
  SDI_F->sah_close_set.finger = finger;
  SDI_F->sah_close_set.data = SDI_F->sah_close[port].fingers[finger];
  return OK;
}


