/*
 * Copyright (c) 2009 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef _SAHANDSTRUCT_H
#define _SAHANDSTRUCT_H

#define SAH_MAX_FINGERS (4)		/* number of fingers */

/* This is used as an index to the arrays representing both hands */
typedef enum SAHAND_PORT {
	SAH_PORT_1 = 0,
	SAH_PORT_2 = 1
} SAHAND_PORT;

typedef enum SAHAND_CONFIG {
	SAH_CONFIG_NOT_PRESENT = 0,
	SAH_CONFIG_LEFT = 1,
	SAH_CONFIG_RIGHT = 2
} SAHAND_CONFIG;

typedef enum SAHAND_BRAKE_STATE {
	SAH_RUN = 0,
	SAH_BRAKE = 1
} SAHAND_BRAKE_STATE;

typedef enum SAHAND_CONTROLLER_TYPE {
	SAH_CONTROLLER_RESET = 0,
	SAH_CONTROLLER_DIRECT = 1,
	SAH_CONTROLLER_POSITION = 2,
	SAH_CONTROLLER_IMPEDANCE = 3
} SAHAND_CONTROLLER_TYPE;

typedef enum SAHAND_FINGER {
	SAH_FINGER_THUMB = 0,
	SAH_FINGER_FIRST = 1,
	SAH_FINGER_MIDDLE = 2,
	SAH_FINGER_RING = 3,
	SAH_FINGER_LITTLE = 4		/* unused */
} SAHAND_FINGER;

/*----------------------------------------------------------------------*/
/*
 * Reference data 
 */

/* refererence for one finger */
typedef struct SAHAND_FINGER_REF {
	double sfr_angle[3];		/* joint angles */
	double sfr_vel[3];		/* velocity in rad/s */
} SAHAND_FINGER_REF;

/* Hand reference */
typedef struct SAHAND_REF {
	double sr_thumb_orientation;
	SAHAND_FINGER_REF sr_fingers[SAH_MAX_FINGERS];
} SAHAND_REF;

/*----------------------------------------------------------------------*/
/*
 * State data
 */
typedef struct SAHAND_THUMB_STATE {
	int sts_brake;
	double sts_angle;
} SAHAND_THUMB_STATE;

/* data for one finger */
typedef struct SAHAND_FINGER_STATE {
	double sfs_pos[3];
	double sfs_vel[3];
	double sfs_torque[3];
	int sfs_limits[3];
	double sfs_finger_temp;
} SAHAND_FINGER_STATE;
	
typedef struct SAHAND_STATE {
	SAHAND_CONFIG ss_config;
	SAHAND_BRAKE_STATE ss_brake;
	double ss_palm_temp;
	SAHAND_THUMB_STATE ss_thumb;
	SAHAND_FINGER_STATE ss_finger[SAH_MAX_FINGERS];
	int ss_error;			/* last error code */
} SAHAND_STATE;

/*----------------------------------------------------------------------*/
/*
 * Parameters
 */

/* stiffness (Nm/degree -- ATTENTION not in radian ) */
typedef struct SAHAND_STIFFNESS_FACTOR {
	double ssf_value[3];
} SAHAND_STIFFNESS_FACTOR;

typedef struct SAHAND_PARAMS {
	SAHAND_CONTROLLER_TYPE sap_controller;  /* controller type */
	SAHAND_STIFFNESS_FACTOR sap_stiffness[SAH_MAX_FINGERS];
} SAHAND_PARAMS;

/*----------------------------------------------------------------------*/
/*
 * Structures to control one particular element
 */

typedef struct SAHAND_THIS_FINGER {
	SAHAND_PORT stf_port;
	SAHAND_FINGER stf_id;
} SAHAND_THIS_FINGER;

typedef struct SAHAND_THIS_CONTROLLER_TYPE {
	SAHAND_PORT stct_port;
	SAHAND_CONTROLLER_TYPE stct_controller;
} SAHAND_THIS_CONTROLLER_TYPE;

/* For one specific finger */
typedef struct SAHAND_THIS_FINGER_REF {
	SAHAND_PORT stfr_port;
	SAHAND_FINGER stfr_id;
	SAHAND_FINGER_REF stfr_finger;
} SAHAND_THIS_FINGER_REF;

typedef struct SAHAND_THIS_THUMB_ANGLE {
	SAHAND_PORT stta_port;
	double stta_angle;
} SAHAND_THIS_THUMB_ANGLE;

typedef struct SAHAND_THIS_FINGER_STIFFNESS {
	SAHAND_PORT stfs_port;
	SAHAND_FINGER stfs_id;
	SAHAND_STIFFNESS_FACTOR stfs_finger;
} SAHAND_THIS_FINGER_STIFFNESS;

typedef struct SAHAND_THIS_HAND_BRAKE {
	SAHAND_PORT sthb_port;
	SAHAND_BRAKE_STATE sthb_brake;
} SAHAND_THIS_HAND_BRAKE;

#define SAHAND_NB_PHALANX 3
typedef enum SAHAND_PHALANX {
  SAHAND_DISTAL = 0,
  SAHAND_PROXIMAL = 1,
  SAHAND_BASE = 2
} SAHAND_PHALANX;

typedef struct SAHAND_FINGER_STIFFNESS {
  double distal;
  double proximal;
  double base;
} SAHAND_FINGER_STIFFNESS;

typedef struct SAHAND_HAND_STIFFNESS {
  SAHAND_PORT hand;
  SAHAND_FINGER_STIFFNESS thumb;
  SAHAND_FINGER_STIFFNESS first;
  SAHAND_FINGER_STIFFNESS middle;
  SAHAND_FINGER_STIFFNESS ring;
} SAHAND_HAND_STIFFNESS;

typedef struct SAHAND_FINGER_VELOCITY {
  double distal;
  double proximal;
  double base;
} SAHAND_FINGER_VELOCITY;

typedef struct SAHAND_HAND_VELOCITY {
  SAHAND_PORT hand;
  SAHAND_FINGER_VELOCITY thumb;
  SAHAND_FINGER_VELOCITY first;
  SAHAND_FINGER_VELOCITY middle;
  SAHAND_FINGER_VELOCITY ring;
  double thumb_orientation;
} SAHAND_HAND_VELOCITY;

typedef struct SAHAND_FINGER_POSITION {
  double distal;
  double proximal;
  double base;
} SAHAND_FINGER_POSITION;

typedef struct SAHAND_HAND_POSITION {
  SAHAND_PORT hand;
  SAHAND_FINGER_POSITION thumb;
  SAHAND_FINGER_POSITION first;
  SAHAND_FINGER_POSITION middle;
  SAHAND_FINGER_POSITION ring;
  double thumb_orientation;
} SAHAND_HAND_POSITION;


typedef enum SAHAND_CLOSE_STATE {
  SAHAND_FAILED           = 0,
  SAHAND_FINGER_INACTIVE  = 1,
  SAHAND_FINGER_CLOSING   = 2, 
  SAHAND_CLOSING_PROXIMAL = 3, 
  SAHAND_CLOSING_DISTAL   = 4,
  SAHAND_TIGHTEN          = 5,
  SAHAND_GRASPING         = 6,
  SAHAND_GRASPED          = 7
} SAHAND_CLOSE_STATE;

typedef struct SAHAND_FINGER_CLOSE {
  double positionEnd[SAHAND_NB_PHALANX];    /* limit position when closing */
  double positionDelta[SAHAND_NB_PHALANX];  /* for tighten */
  double stiffness[SAHAND_NB_PHALANX];      /* stiffness for closing */
  double stiffnessClose[SAHAND_NB_PHALANX]; /* stiffness for tighten */
  double velocity[SAHAND_NB_PHALANX];       /* velocity for closing */
  double velocityClose[SAHAND_NB_PHALANX];  /* velocityClose for tighten */
  double force[SAHAND_NB_PHALANX];          /* force threshold */
  double graspForce;
  double Kpos, Kangle, Kforce;
} SAHAND_FINGER_CLOSE;

typedef struct SAHAND_FINGER_CLOSE_I {
  double force0[SAHAND_NB_PHALANX];
  double positionGrasped[SAHAND_NB_PHALANX];  /* rename ??  closePosition */
  SAHAND_CLOSE_STATE state[SAHAND_NB_PHALANX];
} SAHAND_FINGER_CLOSE_I;

typedef struct SAHAND_HAND_CLOSE {
  SAHAND_FINGER_CLOSE fingers[SAH_MAX_FINGERS];
} SAHAND_HAND_CLOSE;

typedef struct SAHAND_HAND_CLOSE_I {
  SAHAND_FINGER_CLOSE_I fingersI[SAH_MAX_FINGERS];
  SAHAND_CLOSE_STATE stateHand;
  int nbcall;
} SAHAND_HAND_CLOSE_I;

typedef struct SAHAND_CLOSE_FINGER { 
  SAHAND_PORT port;
  SAHAND_FINGER finger;
} SAHAND_CLOSE_FINGER;
 
typedef struct SAHAND_CLOSE_SET {
  SAHAND_PORT port;
  SAHAND_FINGER finger;
  SAHAND_FINGER_CLOSE data;
} SAHAND_CLOSE_SET;

typedef struct SAHAND_TRACK_STR {
  GEN_STRING128       trackPosterName;
  int status;
} SAHAND_TRACK_STR;
  
#endif
